(ns tservice.plugins.merge-rnaseq-expression
  (:require [clojure.data.json :as json]
            [clojure.spec.alpha :as s]
            [clojure.tools.logging :as log]
            [spec-tools.core :as st]
            [clj-yaml.core :as yaml]
            [tservice.lib.files :as ff]
            [tservice.api.storage.fs :as fs-api]
            [tservice.lib.fs :as fs-lib]
            [tservice.vendor.multiqc :as mq]
            [tservice.util :as u]
            [clojure.java.shell :as shell :refer [sh]]
            [tservice.api.task :refer [publish-event! make-plugin-metadata make-events-init create-task! update-process!]]
            [tservice.plugins.merge-rnaseq-expression.merge-exp :as me]
            [tservice.plugins.merge-rnaseq-expression.commons :refer [add-env-to-path correct-filepath]]))

(set! *warn-on-reflection* true)

;;; ------------------------------------------------ Event Specs ------------------------------------------------
(s/def ::enable_multiqc
  (st/spec
   {:spec                boolean?
    :type                :boolean
    :description         "Whether enable to generate multiqc report."
    :swagger/default     true
    :reason              "Only support bollean."}))

(s/def ::excludes
  (st/spec
   {:spec                vector?
    :type                :vector
    :description         "A collection of exclude files."
    :swagger/default     []
    :reason              "The excludes must be an array."}))

(s/def ::filepath
  (st/spec
   {:spec                (s/and string? #(some? (re-matches #"^[a-zA-Z0-9]+:\/\/(\/|\.\/)?[a-zA-Z0-9_]+.*" %)))
    :type                :string
    :description         "File path for covertor."
    :swagger/default     nil
    :reason              "The filepath must be string."}))

(def merge-exp-params-body
  "A spec for the body parameters."
  (s/keys :req-un [::filepath]
          :opt-un [::excludes ::enable_multiqc]))

;;; ------------------------------------------------ Event Metadata -------------------------------------------------

(def metadata
  (make-plugin-metadata
   {:name "merge-rnaseq-exp"
    :summary "Merge expression table for rna-seq."
    :params-schema merge-exp-params-body
    :handler (fn [{:keys [filepath excludes enable_multiqc] :as payload}]
               (let [workdir (ff/get-workdir)
                     from-path (correct-filepath filepath)
                     report-file (fs-lib/join-paths workdir "multiqc_report.html")
                     log-path (fs-lib/join-paths workdir "log")
                     response {:files [report-file
                                       (fs-lib/join-paths workdir "fpkm.csv")
                                       (fs-lib/join-paths workdir "count.csv")]
                               :log log-path
                               :response-type :data2files}
                     task-id (create-task! {:name (str "merge-rnaseq-expression" (u/datetime))
                                            :description "Merge Expression Table for RNA-Seq."
                                            :payload payload
                                            :plugin-name "syncdata"
                                            :plugin-type "ToolPlugin"
                                            :plugin-version "v0.1.0"
                                            :response response})]
                 (fs-lib/create-directories! workdir)
                 ; Launch the mergeexp
                 (spit log-path (json/write-str {:status "Running" :msg ""}))
                 (update-process! task-id 0)
                 (publish-event! "merge_rnaseq_expression"
                                 {:data-dir from-path
                                  :excludes excludes
                                  :enable-multiqc enable_multiqc
                                  :dest-dir workdir
                                  :task-id task-id})
                 response))
    :plugin-type :ToolPlugin
    :response-type :data2files}))

;;; ------------------------------------------------ Event Processing ------------------------------------------------
(defn- decompression-tar
  [filepath]
  (shell/with-sh-env {:PATH   (add-env-to-path "merge-rnaseq-expression")
                      :LC_ALL "en_US.utf-8"
                      :LANG   "en_US.utf-8"}
    (let [command ["bash" "-c"
                   (format "tar -xvf %s -C %s" filepath (fs-lib/parent-path filepath))]
          result  (apply sh command)
          status (if (= (:exit result) 0) "Success" "Error")
          msg (str (:out result) "\n" (:err result))]
      {:status status
       :msg msg})))

(defn- merge-exp!
  [data-dir patterns dest-dir output-file]
  (let [files (fs-api/batch-filter-files data-dir patterns)]
    (log/info "Merge these files: " files)
    (fs-lib/create-directories! dest-dir)
    (log/info "Merge gene experiment files from ballgown directory to a experiment table: " dest-dir output-file)
    (fs-api/copy-files! files dest-dir {:replace-existing true})
    (me/merge-exp-files! (fs-api/list-files dest-dir {:mode "file"}) output-file)))

(defn gen-multiqc-config
  [config-file]
  (let [config (yaml/generate-string
                {:run_modules ["fastqc" "fastq_screen" "qualimap"]}
                :dumper-options {:flow-style :block})]
    (spit config-file config)))

(defn- merge-exp-event!
  [{:keys [data-dir dest-dir excludes enable-multiqc task-id]}]
  ;; TODO: filter excludes?
  (let [log-path (fs-lib/join-paths dest-dir "log")]
    (try
      ;; Merge ballgown files
      (merge-exp! data-dir
                  [".*call-ballgown/.*.txt"]
                  (fs-lib/join-paths dest-dir "ballgown")
                  (fs-lib/join-paths dest-dir "fpkm.csv"))
      (update-process! task-id 30)
      ;; Merge count files
      (merge-exp! data-dir
                  [".*call-count/.*_gene_count_matrix.csv"]
                  (fs-lib/join-paths dest-dir "count")
                  (fs-lib/join-paths dest-dir "count.csv"))
      (update-process! task-id 60)
      (if enable-multiqc
        (let [files (fs-api/batch-filter-files data-dir [".*call-fastqc/.*.zip"
                                                         ".*call-fastqscreen/.*screen.txt"
                                                         ".*call-qualimap/.*bamqc"
                                                         ".*call-qualimap/.*bamqc_qualimap.tar.gz"
                                                         ".*call-qualimap/.*RNAseq"
                                                         ".*call-qualimap/.*RNAseq_qualimap.tar.gz"])
              multiqc-dir (fs-lib/join-paths dest-dir "multiqc")
              config (fs-lib/join-paths dest-dir "multiqc_report.yaml")]
          (fs-lib/create-directories! multiqc-dir)
          (fs-api/copy-files! files multiqc-dir {:replace-existing true})
          (gen-multiqc-config config)
          (doseq [file (fs-api/batch-filter-files multiqc-dir [".*bamqc_qualimap.tar.gz" ".*RNAseq_qualimap.tar.gz"])]
            (decompression-tar file))
          (let [multiqc-result (mq/multiqc multiqc-dir dest-dir {:title "MultiQC Report" :template "default"
                                                                 :config config :env {:PATH (add-env-to-path "merge-rnaseq-expression")}})
                result {:status (:status multiqc-result)
                        :msg (:msg multiqc-result)}
                log (json/write-str result)]
            (log/info "Status: " result)
            (spit log-path log)))
        (spit log-path (json/write-str {:status "Success" :msg ""})))
      (update-process! task-id 100)
      (catch Exception e
        (update-process! task-id -1)
        (let [log (json/write-str {:status "Error" :msg (.toString e)})]
          (log/info "Status: " log)
          (spit log-path log))))))

;;; --------------------------------------------------- Lifecycle ----------------------------------------------------

(def events-init
  "Automatically called during startup; start event listener for merge_rnaseq_expression events."
  (make-events-init "merge_rnaseq_expression" merge-exp-event!))
